#include "hogwash.h"
/*********************************************
* hogwash.c
*
* inline packet scrubber
*********************************************/
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "packet.h"
#include "pass_rule.h"
#include "control.h"
#include "macfilter.h"
#include "defrag.h"

//#define DEBUG
//#define DEBUGPATH

/*------------------Global Structures--------*/
struct PacketHolder{
	char			Buff[PACKET_HOLDER_SIZE];
	struct timeval	timestamp;
	unsigned int 	len;
};

/*------------------Global Variables---------*/
int		no_rules;
int		no_control;
int		no_defrag;
int		spanning_enabled;

struct PacketHolder	Packets[NUM_HOLDERS];

/*------------------Function Declares--------*/

/*********************************************
*  Initialize all variables and stuff
*********************************************/
void InitStructures(){
#ifdef DEBUGPATH
	printf("In InitStructures\n");
#endif

	bzero(internal_name, MAX_INTERFACE_NAME_SIZE);
	bzero(external_name, MAX_INTERFACE_NAME_SIZE);
	bzero(span_name, MAX_INTERFACE_NAME_SIZE);

	no_rules=TRUE;
	no_control=TRUE;
	no_defrag=FALSE;
	spanning_enabled=FALSE;

	bzero(Packets, sizeof(struct PacketHolder) * NUM_HOLDERS);
}

/*********************************************
*  Print the banner so we know who we are
*********************************************/
void ShowBanner(int args){

#ifdef DEBUGPATH
	printf("In ShowBanner\n");
#endif

	printf("hogwash v0.4 by Jason Larsen\n");
	printf("using snort v1.8.1 by Martin Roesch\n\n");
	if (args){
		printf("Usage hogwash <options>\n");
		printf("  -c <rules file>\n");
		printf("  -C <rules file directory> (use highest numbered rules.vXXXXX file)\n");
		printf("  -i <internal interface>\n");
		printf("  -e <external interface>\n");
		printf("  -s <span interface>\n");
		printf("  -l <log directory>\n");
		printf("  -n (no rules mode)\n");
		printf("  -h (show this help message and exit)\n");
		printf("  -d (go to daemon mode)\n");
		printf("  -k <control config file>\n");
		printf("  -f (turn off defragging)\n");
		printf("  --sid <sid file name> (Generate a sid file from the rules file)\n");
		exit(1);
	}
}

/*********************************************
*  Parse the command line args
*********************************************/
void ParseArgs(int argc, char* argv[]){
	int	i=0;
	
#ifdef DEBUGPATH
	printf("In ParseArgs\n");
#endif
		
	if (argc==1) ShowBanner(TRUE);
	
	for (i=0; i<argc; i++){	
		if (strncasecmp("-i",argv[i],2)==0){
			if ( (i>=argc-1) || (argv[i+1][0]=="-"[0])){
				printf("Syntax -i <internal interface>\n");
				exit(1);
			}
			snprintf(internal_name, MAX_INTERFACE_NAME_SIZE, argv[i+1]);
#ifdef DEBUG			
			printf("New Internal %s\n",internal_name);
#endif			
			i++;
		}else if (strncasecmp("-e",argv[i],2)==0){
			if ( (i>=argc-1) || (argv[i+1][0]=="-"[0]) ){
				printf("Syntax -e <external interface>\n");
				exit(1);
			}
			snprintf(external_name, MAX_INTERFACE_NAME_SIZE, argv[i+1]);
#ifdef DEBUG			
			printf("New External %s\n",external_name);
#endif			
		}else if (strncasecmp("-s",argv[i],2)==0){
			if ( (i>=argc-1) || (argv[i+1][0]=="-"[0]) ){
				printf("Syntax -s <span interface>\n");
				exit(1);
			}
			spanning_enabled=TRUE;
			snprintf(span_name, MAX_INTERFACE_NAME_SIZE, argv[i+1]);
#ifdef DEBUG
			printf("New Span %s\n",span_name);
#endif
		}else if (strncasecmp("-n", argv[i],2)==0){
			no_rules=TRUE;
			printf("******************************************\n");
			printf("* Entering No Rules Mode\n");
			printf("******************************************\n");
		}else if (strncasecmp("-f", argv[i],2)==0){
			no_defrag=TRUE;
#ifdef DEBUG
			printf("Defragging turned off\n");
#endif			
		}else if (strncasecmp("-h",argv[i],2)==0){
			ShowBanner(TRUE);
			exit(1);
		}
		else if (strncasecmp("-k",argv[i],2)==0)
		{
			no_control=FALSE;
		}
		else if (strncasecmp("-c",argv[i],2)==0)
		{
			no_rules=FALSE;
		}
		else if (strncasecmp("-d",argv[i],2)==0)
		{
			printf("Going into daemon mode\n");
			daemon(0,0);
		}

	}
}

/****************************************************
* Creates a PID so we don't run 2
****************************************************/
void CreateLock(){
	int			fp;
	struct stat	s;
	int			pid;
	int			statv;
	char		Buff[1024];
	
#ifdef DEBUGPATH
	printf("In CreateLock\n");
#endif
		
	bzero(&s, sizeof(struct stat));
	statv=stat(PID_FILE, &s);
	if (statv!=0){
		fp = open(PID_FILE, O_CREAT|O_RDWR);
		if (!fp){
			printf("Couldn't create PID file %s.\n",PID_FILE);
			exit(1);
		}
		pid=getpid();
		sprintf(Buff,"%i\n",pid);
		write(fp, Buff, strlen(Buff));
		close(fp);
		return;
	}
	
	printf("PID file %s exists. Running two copies of hogwash may crash some switches.\n", PID_FILE);  
	exit(1);
	
}

/*********************************************
*  Clear the lock
*********************************************/
void DestroyLock(int signum){

#ifdef DEBUGPATH
	printf("In DestroyLock\n");
#endif

	printf("Exiting Hogwash with signal %i..\n",signum);
	remove(PID_FILE);
	pass_rule_free();
	close_interfaces(spanning_enabled);
	exit(0);
}

/*********************************************
*  Catch signals and deal with them
*********************************************/
void BindSignals(){

#ifdef DEBUGPATH
	printf("In BindSignals\n");
#endif

	signal(SIGTERM, DestroyLock);
	signal(SIGINT,  DestroyLock);
	signal(SIGQUIT, DestroyLock);
	signal(SIGHUP,  restart);
}

/*********************************************
*  Packet processing starts here
*********************************************/
void HogwashProcessPacket(PacketToProcess* ptp, int spanning){
	int	from_interface;
	int to_interface;
	
#ifdef DEBUG
	printf("Processing Packet of len %u from interface %i\n", ptp->packet_len, ptp->from_interface);
#endif	

	from_interface=ptp->from_interface;
	if (from_interface==INTERNAL) to_interface=EXTERNAL; else to_interface=INTERNAL;
	
	if (ProcessMac(ptp)){
#ifdef DEBUG	
		printf("Mac filter dropped the packet\n");
#endif		
		return;
	}
	
	if (!no_defrag)
	if (ProcessFrag(ptp)) return;
	
	if (!no_rules)
	if (pass_rule(ptp, spanning)) return;
	
	if (!no_control)
	if (control_func(ptp, spanning)==RESPONSE_DROP) return;
	
#ifdef DEBUG
	printf("writing packet\n");
#endif	
	write_packet(to_interface, ptp->raw_packet, ptp->packet_len, spanning);
}

/*********************************************
*  The one and only main
*********************************************/
int main(int argc, char**argv){

#ifdef DEBUGPATH
	printf("In main\n");
#endif

	InitStructures();
	
	ParseArgs(argc, argv);
	CreateLock();
	if (!no_rules) pass_rule_init(argc, argv);
	if (!no_control) control_init(argc, argv);
	if (!no_defrag) FragInit(argc, argv);
	mac_init(argc,argv);
	
	BindSignals();
	open_interfaces(internal_name, external_name, span_name, spanning_enabled);
	main_loop(spanning_enabled);
	DestroyLock(0);
	
	return TRUE;
}
