#include "packet.h"
#include "hogwash.h"

#ifdef __FREEBSD__
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <time.h>
#include <net/bpf.h>
#include <errno.h>
#include <poll.h>
#include <err.h>
#include <string.h>
#include <stdlib.h>

#define FATAL		1
#define NONFATAL	0

#include <stdio.h>
int pcount;

/***********************************************
* Globals
***********************************************/
int fd1, fd2;

/* SPAN */
int fd3;

/***********************************************
* Open up a bpf
***********************************************/
static int
bpf_open()
{
	int     fd;
	int     n = 0;
	char    device[sizeof "/dev/bpf0000000000"];

	/* Go through all the minors and find one that isn't in use. */
	do {
		(void) snprintf(device, sizeof device, "/dev/bpf%d", n++);
		fd = open(device, O_RDWR);
	} while (fd < 0 && errno == EBUSY);

	if (fd < 0) {
		err(FATAL, "%s: %s", device, strerror(errno));
		/* NOTREACHED */
	}
	return fd;
}


/*
 * Open a BPF file and attach it to the interface named 'device'.
 * Set immediate mode, and set a filter that accepts only RARP requests.
 */
int open_interface(char*device){
	int     fd;
	struct ifreq ifr;
	u_int   dlt;
	int     immediate;
	int	promisc;
	int	link;
	int	loopback;

	fd = bpf_open();

	/* Set immediate mode so packets are processed as they arrive. */
	immediate = 1;
	if (ioctl(fd, BIOCIMMEDIATE, &immediate) < 0) {
		err(FATAL, "BIOCIMMEDIATE: %s", strerror(errno));
		/* NOTREACHED */
	}

	(void) strncpy(ifr.ifr_name, device, sizeof ifr.ifr_name);
	if (ioctl(fd, BIOCSETIF, (caddr_t) & ifr) < 0) {
		err(FATAL, "BIOCSETIF: %s", strerror(errno));
		/* NOTREACHED */
	}
	/* Check that the data link layer is an Ethernet; this code won't work
	 * with anything else. */
	if (ioctl(fd, BIOCGDLT, (caddr_t) & dlt) < 0) {
		err(FATAL, "BIOCGDLT: %s", strerror(errno));
		/* NOTREACHED */
	}
	if (dlt != DLT_EN10MB) {
		err(FATAL, "%s is not an ethernet", device);
		/* NOTREACHED */
	}
	/*turn off automatic filling of the mac address*/
	link=1;
	if (ioctl(fd, BIOCSHDRCMPLT, &link) < 0){
		printf("Couldn't turn off auto mac address\n");
		exit(0);
	}
	loopback=0;
	if (ioctl(fd, BIOCGSEESENT, &loopback) < 0){
		printf("Couldn't set loopback\n");
		exit(0);
	}
	/* Switch the interface into promisc mode*/
	promisc=1;
	if (ioctl(fd, BIOCPROMISC, &promisc) < 0) {
		printf("Couldn't enter promisc mode\n");
		exit(1);
	}
	if (dlt != DLT_EN10MB) {
		err(FATAL, "%s is not an ethernet", device);
		/* NOTREACHED */
	}	

	return fd;
}

/***********************************************
* Open the interfaces for reading and writing
***********************************************/
int open_interfaces(char* internalif, char* externalif, char* spanif, int use_span){
	pcount=0;
	fd1=open_interface(internalif);
	fd2=open_interface(externalif);
	if (use_span == TRUE) {
	  fd3=open_interface(spanif);
	  return (fd1>0 && fd2>0 && fd3>0);
	} else {
	  return (fd1>0 && fd2>0);
	}
}

/***********************************************
* close the interfaces for reading and writing
***********************************************/
void close_interfaces(int use_span){
	close(fd1);
	close(fd2);
	if (use_span == TRUE) {
	  close(fd3);
	}
}

/***********************************************
* Get packets
***********************************************/
void main_loop(int use_span){
	int 			bufsize1, bufsize2, cc;
	u_char			*buf1, *buf2, *bp, *ep;
	int				i;
	PacketToProcess	ptp;
	
	struct timeval	tv;	
	struct pollfd	pds[2];
	int				numfds;
	
	/*initialize the set*/
	pds[0].fd=fd1;
	pds[0].events=POLLIN;
	pds[0].revents=0;
	pds[1].fd=fd2;
	pds[1].events=POLLIN;
	pds[1].revents=0;

	/*Get the sizes of the buffers*/
	if (ioctl(fd1, BIOCGBLEN, (caddr_t) & bufsize1) < 0) {
		err(FATAL, "BIOCGBLEN: %s", strerror(errno));
		/* NOTREACHED */
	}
	printf("Buffer size1 is %i\n",bufsize1);
	buf1 = (u_char *) malloc((unsigned) bufsize1);
	if (buf1 == 0) {
		err(FATAL, "malloc: %s", strerror(errno));
		/* NOTREACHED */
	}
	
	if (ioctl(fd2, BIOCGBLEN, (caddr_t) & bufsize2) < 0) {
		err(FATAL, "BIOCGBLEN: %s", strerror(errno));
		/* NOTREACHED */
	}
	printf("Buffer size2 is %i\n",bufsize2);
	buf2 = (u_char *) malloc((unsigned) bufsize2);
	if (buf2 == 0) {
		err(FATAL, "malloc: %s", strerror(errno));
		/* NOTREACHED */
	}
	/*end getting the sizes of the buffers*/
	
	while (1){
		tv.tv_sec=1;
		tv.tv_usec=0;
		numfds=poll(pds, 2, -1);
		
		if (numfds>0){
			for (i=0;i<2;i++){
				if ( (pds[i].revents & POLLIN)== POLLIN ){
#ifdef DEBUG
					printf("__________________________\n");
					printf("Got packet on interface %i\n", i);
#endif
					again:
					bzero(buf1,bufsize1);
					cc = read(pds[i].fd, (char *) buf1, bufsize1);
#ifdef DEBUG				
					printf("Read %i bytes from the bpf\n",cc);
#endif					
					/* Don't choke when we get ptraced */
					if (cc < 0 && errno == EINTR)
						goto again;
					if (cc < 0) {
						if (errno == EINVAL &&
					    	(lseek(pds[i].fd, 0, SEEK_CUR) + bufsize1) < 0) {
							(void) lseek(pds[i].fd, 0, 0);
							goto again;
						}
						err(FATAL, "read: %s", strerror(errno));
						/* NOTREACHED */
					}
					/* Loop through the packet(s) */
#define bhp ((struct bpf_hdr *)bp)
					bp = buf1;
					ep = bp + cc;
					while (bp < ep) {
						register int caplen, hdrlen;

						caplen = bhp->bh_caplen;
						hdrlen = bhp->bh_hdrlen;
						bzero(&ptp,sizeof(PacketToProcess));
#ifdef DEBUG
						printf("This packet is %u bytes\n",bhp->bh_caplen);
#endif						
						ptp.packet_len=bhp->bh_caplen;
						ptp.capture_len=bhp->bh_caplen;
						ptp.ts.tv_sec=bhp->bh_tstamp.tv_sec;
						ptp.ts.tv_usec=bhp->bh_tstamp.tv_usec;
						memcpy(ptp.raw_packet, bp+18, ptp.packet_len);
						ptp.from_interface=i;
						HogwashProcessPacket(&ptp, use_span);
						bp += BPF_WORDALIGN(hdrlen + caplen);
					}
				}
			}
		}
	}
}

/***********************************************
* Restart Hogwash
***********************************************/
void restart(){
}

/***********************************************
* Send out a packet
***********************************************/
int write_packet(int interface, char* packet, int packet_len, int use_span){
	int cc;
	int	fd;
	struct	pollfd	pfd;
#ifdef LOG_WRITES	
	FILE*	fp;
	char	fname[1024];
#endif	

#ifdef LOG_WRITES
	if (interface==0){
		snprintf(fname, 1024, "%i.packet0", pcount);
		fp=fopen(fname, "w");	
		fwrite(packet, packet_len, 1, fp);
		fclose(fp);	
		pcount++;
	}
#endif
	
#ifdef DEBUG
	printf("Writing packet OBSD to interface %i\n",interface);
#endif

	if (interface==INTERNAL)
		fd=fd1;
	else
		fd=fd2;

	/*zot test*/
	pfd.fd=fd;
	pfd.events=POLLOUT;
	poll(&pfd, 1, 11);
	/*end zot test*/

        if (use_span == TRUE) {
	  cc=write(fd3, packet, packet_len);
	  if (cc!=packet_len){
#ifdef DEBUG
		printf("write_packet: span packet copy size did not match\n");
#endif
	  }
	}

	cc=write(fd, packet, packet_len);
	//usleep(11);
	if (cc!=packet_len){
#ifdef DEBUG	
		printf("write_packet: sizes did not match\n");
#endif			
	//	return FALSE;
	}
	return TRUE;
}
#endif
