/*
** Copyright (C) 1998,1999,2000,2001 Martin Roesch <roesch@clark.net>
** Copyright (C) 2001 Jed Haile  <jhaile@nitrodata.com>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/* state based portscan detector
*  by Jed Haile <jhaile@nitrodata.com>
*  version 0.0.1
*  todo:  1. track timestamp, src, dst, proto, sport/icode, dport/itype, length 
*/

#define MODNAME "spp_scanmunge"
#define MAX_TARGETS 5
#define MAX_PORTS 20
#define PS_T 1
#define TGT_T 2
#define TRUE 1
#define FALSE 0

#include "spp_scanmunge.h"
#include "spp_conversation.h"
#include "rules.h"
#include "decode.h"
#include "ubi_SplayTree.h"
#include "ubi_BinTree.h"
#include <netinet/in.h>
#include "generators.h"


typedef struct _spp_timeval
{
	time_t tv_sec;
	time_t tv_usec;
} spp_timeval;

/***************************************************************
* struct to track timestamp, src, dst, proto, sport/icode,     *
* dport/itype, length for each packet from a potential scanner *
****************************************************************/
typedef struct _scannerStat
{	
	spp_timeval timestamp;
	struct in_addr src;
	struct in_addr dst;
	u_int8_t proto;
	u_int16_t sport;
	u_int16_t dport;
	u_int16_t len;
	//struct _scannerStat *prev;
	struct _scannerStat *next;
} scannerStat;


typedef struct _PortScanner
{
	ubi_trNode Node;             /* for the splay tree */
	struct _PortScanner *next;   /* for the node pool */
	struct _PortScanner *prev;   
	struct _PortScanner *pnext;  /* for the pruneList */
	struct _PortScanner *pprev;
    
	u_int32_t scanner_ip;
	unsigned long ssnid;   /* from spp_conversation */

	spp_timeval initial_time;
	spp_timeval last_update_time;

	int totalPrtcount;  /* total count of ports this scanner has hit */
	int totalTgtcount;  /* total count of targets this scanner has hit */
    
	ubi_trRoot targetRoot;  /* this scanner's target tree*/
	ubi_trRootPtr targetRootPtr; 

	int targetsExceeded;
	int portsExceeded;
	int eid;
	scannerStat *stat;

} Portscanner;

typedef struct _Target
{
	ubi_trNode Node;
	struct _Target *next; /* for the node pool */
	struct _Target *prev;
	Portscanner *parent;  /*pointer to the parent Portscanner node */
    
	u_int32_t target_ip;
	int prtcount;      /* number of ports on this target hit by parent portscanner*/
    
	spp_timeval initial_time;
	spp_timeval last_update_time;

	char plist[65536/8];  /* an array of bytes to store port info */

} Target;


typedef struct _scanData
{
	char *logpath;
	int psnodes;
	int tgtnodes;
	int tgtThreshold;
	int portThreshold;
	spp_timeval timeout;
} ScanData;

typedef struct _plist_t 
{
	Portscanner nilnode;
	int nodecount;
	int maxcount;
} plist_t;

typedef struct _PortscannerNodePool
{
	Portscanner nil;
	int size;
	int used;
} PortscannerNodePool;

typedef struct _TargetNodePool
{
	Target nil;
	int size;
	int used;
} TargetNodePool;

typedef struct _hostNode
{
	IpAddrSet *address;
	char ignoreFlags;
	struct _hostNode *nextNode;
}	hostNode;

/** SPLAY TREE ROOT DATA (and other globals)**/
ubi_trRoot ps_cache;
ubi_trRootPtr psRootPtr = &ps_cache;
plist_t *pruneList; 
static ubi_trRoot sessionRoot;
static ubi_trRootPtr sessionRootPtr = &sessionRoot;
PortscannerNodePool* psPoolPtr;
TargetNodePool* tgtPoolPtr;

/** Global Program Data **/
FILE *scanlog; 
hostNode *hostList; //for ignore-hosts
psData mdata;
ScanData sdata;

/** external globals from rules.c **/
extern char *file_name;
extern int file_line;
extern u_int32_t event_id;



/** FUNCTION PROTOTYPES **/
void Scan2Init(u_char *);
static int targetCompareFunc(ubi_trItemPtr , ubi_trNodePtr);
//void AddTarget(ubi_trRootPtr, u_int32_t, u_int16_t, u_int32_t);
void TruncateScanList(plist_t *, int);
void ParseScanmungeArgs(u_char *);
void SLog(Packet *, int, Portscanner *);
void SAlert(Packet *, int, Portscanner *);
inline int portIsSet(char *, int);
void setPort(char *, int);
void dumpPacketStats(Portscanner *);

/* For node managment */
Portscanner* Portscanner_borrow();
Target* Target_borrow();
void Portscanner_return(Portscanner *);
void Target_return(Target *);

/* For ignore hosts */
void SetupScanIgnoreHosts();
void InitIgnoreHosts(u_char *);
IpAddrSet* IgnoreAllocAddrNode(hostNode * host);
void ScanParseIp(char *, hostNode *);
int IsIgnored(Packet *);

/* For pruneList management */
int list_isempty(plist_t *);
Portscanner* list_last(plist_t *);
Portscanner* list_delete(plist_t *, Portscanner *);
plist_t* list_create(int);
int list_contains(plist_t *, Portscanner *);
void list_prepend(plist_t *, Portscanner *);
int list_isfirst(plist_t *, Portscanner *);

extern int FilterThis;

/*************************************************************/
/* Parses all of scan2's args. They are as follows:          */
/* int psnodes, int targetnodes, char log, int targets,      */
/* int ports, int timeout                                    */
/*************************************************************/
void ParseScanmungeArgs(u_char *args)
{
	int num_toks, s_toks;
	char **toks;
	char **stoks;
	int i;
	char* index;
	char logpath[STD_BUF], tmp[STD_BUF];
	
	if (args==NULL){
		strncpy(logpath, pv.log_dir, STD_BUF);
		strncpy(tmp, "/scan.log", STD_BUF);
		strncat(logpath, tmp, STD_BUF);
		scanlog = fopen(logpath, "a+");
		sdata.psnodes = 100;
		sdata.tgtnodes = 250;
		sdata.tgtThreshold = 5;
		sdata.portThreshold = 20;
		sdata.timeout.tv_sec = 60;
		return;
	}

	toks = mSplit(args, ",", 11, &num_toks, 0);

	i=0;

	while (i < num_toks)
	{
		index = toks[i];

		while(isspace((int)*index)) index++;

		stoks = mSplit(index, " ", 4, &s_toks, 0);

		if(!strncasecmp(stoks[0], "psnodes", 7))
		{
			if(isdigit((int)stoks[1][0]))
			{
			// number of psnodes
				sdata.psnodes = (int)stoks[1][0];
				i++;
			}
			else
			{
				//error
				printf("Error in argument psnodes, please specify a positive integer\n");
				exit(-1);
			}
		}
		else if(!strncasecmp(stoks[0], "tgtnodes", 8))
		{
			if(isdigit((int)(stoks[1][0])))
			{
			// number of tgtnodes
				sdata.tgtnodes = (int)stoks[1][0];
				i++;
			}
			else
			{
			//error
				printf("Error in argument tgtnodes, please specify a positive integer\n");
				exit(-1);
			}
		}
		else if(!strncasecmp(stoks[0], "targets", 7))
		{
			if(isdigit((int)(stoks[1][0])))
			{
			// number of targets
				sdata.tgtThreshold = (int)stoks[1][0];
				i++;
			}
			else
			{
			//error
				printf("Error in argument targets, please specify a positive integer\n");
				exit(-1);
			}
		}
		else if(!strncasecmp(stoks[0], "ports", 7))
		{
			if(isdigit((int)(stoks[1][0])))
			{
			// number of ports
				sdata.portThreshold = (int)stoks[1][0];
				i++;
			}
			else
			{
			//error
				printf("Error in argument ports, please specify a positive integer\n");
				exit(-1);
			}
		}
		else if(!strncasecmp(stoks[0], "timeout", 7))
		{
			if(isdigit((int)(stoks[1][0])))
			{
			// number of ports
				sdata.timeout.tv_sec = (int)stoks[1][0];
				i++;
			}
			else
			{
			//error
				printf("Error in argument ports, please specify a positive integer\n");
				exit(-1);
			}
		}
		else if(!strncasecmp(stoks[0], "log", 3))
		{
			if(isascii((int)(stoks[1][0])))
			{
				char logpath[STD_BUF], tmp[STD_BUF];
				strncpy(logpath, pv.log_dir, STD_BUF);
				strncpy(tmp, "/scan.log", STD_BUF);
				strncat(logpath, tmp, STD_BUF);
				scanlog = fopen(logpath, "a+");
				i++;
			}
			else
			{
			//error
				printf("Error in argument log, please specify a valid log dir\n");
				exit(-1);
			}
		}
	}		/* TODO Free all those toks */
	if(scanlog == NULL)
	{	
		strncpy(logpath, pv.log_dir, STD_BUF);
		strncpy(tmp, "/scan.log", STD_BUF);
		strncat(logpath, tmp, STD_BUF);
		scanlog = fopen(logpath, "a+");
	}
	if (sdata.psnodes == 0)
	{
		sdata.psnodes = 150;
	}
	if (sdata.tgtnodes == 0)
	{
		sdata.tgtnodes = 250;
	}
	if (sdata.portThreshold == 0)
	{
		sdata.portThreshold = 20;
	}
	if (sdata.tgtThreshold == 0)
	{
		sdata.tgtThreshold = 5;
	}
	if (sdata.tgtThreshold == 0)
	{
		sdata.timeout.tv_sec = 60;
	}
}

/************************************************************/
/* Registers scan2-ignorehosts preproc with snort.          */
/************************************************************/
void SetupScanIgnoreHosts(void)
{
	RegisterPreprocessor("scan2-ignorehosts", InitIgnoreHosts);
}

/*************************************************************/
/* Called at runtime to establish the list of hosts who are  */
/* to be ignored by the portscan detector                    */
/*************************************************************/
void InitIgnoreHosts(u_char *hosts)
{
	char **toks;
	int num_toks;
	int num_hosts = 0;
	hostNode *currentHost;
	int i;

	currentHost = NULL;
	hostList = NULL;
	
	if(hosts == NULL)
	{
		FatalError(MODNAME ": ERROR: %s  (%d)=> No arguments to scanmunge-ignorehosts.\n", file_name, file_line);
	}

	toks = mSplit(hosts, " ", 31, &num_toks, '\\');
	
	for(num_hosts = 0; num_hosts < num_toks; num_hosts++)
	{
        if(currentHost != NULL)
        {
            currentHost->nextNode = (hostNode *) calloc(sizeof(hostNode), sizeof(char));
            currentHost = currentHost->nextNode;
        }
        else
        {
            currentHost = (hostNode *) calloc(sizeof(hostNode), sizeof(char));
            hostList = currentHost;
        }

#ifdef DEBUG
        printf(MODNAME ": CreateServerList(): Adding server %s\n", toks[num_servers]);
#endif  //DEBUG

        ScanParseIp(toks[num_hosts], currentHost);

#ifdef DEBUG
        memset(ruleIP, '\0', 16);
        memset(ruleNetMask, '\0', 16);
        strncpy(ruleIP, inet_ntoa(*(struct in_addr *) & currentHost->address->ip_addr), 15);
        strncpy(ruleNetMask, inet_ntoa(*(struct in_addr *) & currentHost->address->netmask), 15);
        printf(MODNAME ": CreateServerList(): Added server %s/%s\n", ruleIP, ruleNetMask);
#endif  //DEBUG

        currentHost->nextNode = NULL;
    }

    for(i = 0; i < num_toks; i++)
    {
        free(toks[i]);

	}
}

/************************************************************/
/* Helper function to set up the list of ignored hosts      */
/************************************************************/
IpAddrSet* IgnoreAllocAddrNode(hostNode * host)
{
	IpAddrSet *idx;

    if(host->address == NULL)
    {
        host->address = (IpAddrSet *) calloc(sizeof(IpAddrSet), sizeof(char));

        if(host->address == NULL)
        {
            FatalError("[!] ERROR: Unable to allocate space for portscan IP addr\n");
        }
        return host->address;
    }
    idx = host->address;

    while(idx->next != NULL)
    {
        idx = idx->next;
    }

    idx->next = (IpAddrSet *) calloc(sizeof(IpAddrSet), sizeof(char));

    idx = idx->next;

    if(idx == NULL)
    {
        FatalError("[!] ERROR: Unable to allocate space for portscan IP address\n");
    }

	return idx;
}

/*******************************************************************/
/* parses the IP's in the ignore hosts list                        */
/*******************************************************************/
void ScanParseIp(char *addr, hostNode *host)
{
    char **toks;
    int num_toks;
    int i;
    IpAddrSet *tmp_addr;
    char *tmp;
	char homeFlags;
 
	if(addr == NULL)
	{
	        FatalError("ERROR %s(%d) => Undefined address in portscan-ignorehosts directive\n", file_name, file_line);
    }

    if(*addr == '!')
    {
        homeFlags |= EXCEPT_DST_IP;
        addr++;
    }
 
    if(*addr == '$')
    {
        if((tmp = VarGet(addr + 1)) == NULL)
        {
            FatalError("ERROR %s (%d) => Undefined variable %s\n", file_name, file_line, addr);
        }
    }
    else
    {
        tmp = addr;
    }
 
    if (*tmp == '[')
    {
        *(strrchr(tmp, (int)']')) = 0; /* null out the en-bracket */
 
        toks = mSplit(tmp+1, ",", 128, &num_toks, 0);
 
        for(i = 0; i < num_toks; i++)
        {
            tmp_addr = IgnoreAllocAddrNode(host);
 
            ParseIP(toks[i], tmp_addr);
        }
		 for(i=0;i<num_toks;i++)
            free(toks[i]);
    }
    else
    {
        tmp_addr = IgnoreAllocAddrNode(host);

        ParseIP(tmp, tmp_addr);
    }
}

/************************************************************/
/* checks to see if a packet is coming from an ignored host */
/************************************************************/
int IsIgnored(Packet *p)
{
#ifdef DEBUG
	char sourceIP[16], ruleIP[16], ruleNetMask[16];
#endif
	hostNode *currentHost = hostList;
	
    while(currentHost)
    {
        /*
         * Return 1 if the source addr is in the serverlist, 0 if nothing is
         * found.
         */
        if(CheckAddrPort(currentHost->address, 0, 0, p,
                         (ANY_SRC_PORT | currentHost->ignoreFlags), CHECK_SRC))
        {

#ifdef DEBUG
            memset(sourceIP, '\0', 16);
            memset(ruleIP, '\0', 16);
            memset(ruleNetMask, '\0', 16);
            strncpy(sourceIP, inet_ntoa(p->iph->ip_src), 15);
            strncpy(ruleIP, inet_ntoa(*(struct in_addr *) & (currentHost->address->ip_addr)), 14);
            strncpy(ruleNetMask, inet_ntoa(*(struct in_addr *) & (currentHost->address->netmask)), 15);

            printf(MODNAME ": IsServer():  Server %s found in %s/%s!\n", sourceIP, ruleIP, ruleNetMask);
#endif  //DEBUG
            return(1);
        }
        currentHost = currentHost->nextNode;
    }

    return(0);
}

/*****************************************************************/
/* initialize a pool of either portscanner nodes or target nodes */
/*****************************************************************/
void nodepool_init(int count, int type)
{
	int i;
	Portscanner *ps = NULL;
	Target *tgt;

		if(type==PS_T)
		{
			psPoolPtr = malloc(sizeof *psPoolPtr);
			if (psPoolPtr) 
			{
				/* set up the initial node */
				psPoolPtr->nil.next = &psPoolPtr->nil;
				psPoolPtr->nil.prev = &psPoolPtr->nil;
				psPoolPtr->used = 0;
				psPoolPtr->size = count;
			}
			/* malloc the rest of the nodes and stick them in the pool */
			for (i=0; i < count; i++)
			{
				ps = malloc(sizeof(Portscanner));
				bzero(ps, sizeof(Portscanner));
				ps->stat = malloc(sizeof(scannerStat));
				bzero(ps->stat, sizeof(scannerStat));
				ps->stat->next = NULL;
				//ps->stat->next = ps->stat;
				//ps->stat->prev = ps->stat;
				if (ps)
					Portscanner_return(ps);
				else
				{
					printf("Failed to malloc Portscanner!!!\n");
					exit(-1);
				}
			}
		}
		if(type==TGT_T)
		{
			tgtPoolPtr = malloc(sizeof *tgtPoolPtr);
			//printf("sizeof tgtPoolPtr = %u\n", sizeof *tgtPoolPtr);
			if (tgtPoolPtr) 
			{
				/* set up the initial node */
				tgtPoolPtr->nil.next = &tgtPoolPtr->nil;
				tgtPoolPtr->nil.prev = &tgtPoolPtr->nil;
				tgtPoolPtr->used = 0;
				tgtPoolPtr->size = count;
			}
			/* malloc the rest of the nodes and stick them in the pool */
			for (i=0; i < count; i++)
			{
				tgt = malloc(sizeof(Target));
				bzero(tgt, sizeof(Target));
				if (tgt)
					Target_return(tgt);
				else
				{
					printf("Failed to malloc Portscanner!!!\n");
					exit(-1);
				}
			}
		}
}

/*************************************************/
/* returns a node to use in the portscanner tree */
/*************************************************/
Portscanner* Portscanner_borrow()
{
	Portscanner *new;
	Portscanner *next, *prev;

	if (psPoolPtr->nil.next == &psPoolPtr->nil)
	{
		/* no free nodes, so lop the oldest off the list */
#ifdef DEBUG
		printf("Calling Truncate from Portscanner_borrow\n");
#endif
		TruncateScanList(pruneList, 1);
	}
	/* grab a node off the top */
	new = (Portscanner *)psPoolPtr->nil.next;
	/* link nodes back together */
	next = new->next;
	prev = new->prev;
	prev->next = next;
	next->prev = prev;
	psPoolPtr->used++;
	/* clean up the node */
	new->next = new->prev = NULL;
	new->Node.Link[0] = NULL;
	new->Node.Link[1] = NULL;
	new->Node.Link[2] = NULL;
	return new;
}

/*************************************************/
/* borrows a node from the target node pool for  */
/* use in a target tree                          */
/*************************************************/
Target* Target_borrow()
{
	Target *new;
	Target *next, *prev;

	if (tgtPoolPtr->nil.next == &tgtPoolPtr->nil)
	{
		/* no free nodes, so lop the oldest off the list */
#ifdef DEBUG
		printf("Calling Truncate from Portscanner_borrow\n");
#endif
		TruncateScanList(pruneList, 1);
	}
	/* grab a node off the top */
	new = (Target *)tgtPoolPtr->nil.next;
	/* link nodes back together */
	next = new->next;
	prev = new->prev;
	prev->next = next;
	next->prev = prev;
	tgtPoolPtr->used++;
	/* clean up the node */
	new->next = new->prev = NULL;
	new->Node.Link[0] = NULL;
	new->Node.Link[1] = NULL;
	new->Node.Link[2] = NULL;
	return new;
}

/*************************************************/
/* returns a node to the portscanner node pool   */
/*************************************************/
void Portscanner_return(Portscanner *node)
{
	Portscanner *nil = &psPoolPtr->nil;
	Portscanner *nilnext = nil->next;
	scannerStat *tmp, *tmp1;

	/* the order looks like this:
	*	nil->nilnext  becomes
	*	nil->node->nilnext
	*/
	
	/* First free the stats list attached to this portscanner, 
	 * leaving only the first one */
	if (node->stat != node->stat->next)
	{
		tmp = node->stat->next;
		//free(node->stat);
		while (tmp != NULL)
		{
			tmp1 = tmp->next;
			free(tmp);
			tmp = tmp1;
		}
		node->stat->next = NULL;
		//node->stat->prev = node->stat;
	}

	/* now stick the node back into the node pool's linked list */
	node->prev = nil;     /* stick node right after nil node */
	node->next = nilnext; 
	nilnext->prev = node; 
	nil->next = node;
	psPoolPtr->used--;
}

/*************************************************/
/* returns a node to the target node pool        */
/*************************************************/
void Target_return(Target *node)
{
	Target *nil = &tgtPoolPtr->nil;
	Target *nilnext = nil->next;

	/* the order looks like this:
	*	nil->nilnext  becomes
	*	nil->node->nilnext
	*/
	node->prev = nil;     /* stick node right after nil node */
	node->next = nilnext; 
	nilnext->prev = node; 
	nil->next = node;
	tgtPoolPtr->used--;
}

/************************************************/
/* wrapper function to init the target and      */
/* portscan node pools.                         */
/************************************************/
void InitNodePools(int numps, int numtgt)
{
	nodepool_init(numps, PS_T);
	nodepool_init(numtgt, TGT_T);
}

/********************************************************/
/* takes a target node and zeros out it's port list     */
/********************************************************/
void Initportlist(Target *target)
{
	int i;
	
	for(i=0; i<65536/8; i++)
	{
		target->plist[i] = 0;
	}
}

/**********************************************************/
/* moves the Portscanner ps to the front of the pruneList */
/* The pruneList is ordered by time last active.          */
/**********************************************************/
void UpdatePruneList(Portscanner *ps)
{
	if (list_isfirst(pruneList, ps))
		return;
	if ((list_contains(pruneList, ps)) & (pruneList->nodecount>1)){
		ps = list_delete(pruneList, ps);
	}

	list_prepend(pruneList, ps);
}

/*****************************************************************/
/* Callback function used to delete a target splay tree's nodes  */
/*****************************************************************/
void KillTargetSpd(ubi_trRootPtr NodePtr) 
{
	Target *tmp;
	tmp = (Target *)NodePtr;
	Target_return(tmp);
}

/***************************************************************/
/* Deletes a given target splay tree, freeing/returning all    */
/* resources                                                   */
/***************************************************************/
void DeleteTargetSpd(ubi_trRootPtr RootPtr)
{
	(void)ubi_trKillTree(RootPtr, KillTargetSpd);
}
	
/***************************************************/
/* Add a port # to the port array for the target.  */
/* This is called whenever a portscanner touches   */
/* a new port on a target.                         */
/***************************************************/
inline void AddPort(Target *target, u_int16_t offset, Packet *p)
{
		/* target->plist is an array of char being treated as */
		/* a bitfield. There 65535 bits in the char array.    */
		/* Through a little voodoo we can set any particular  */
		/* bit in that field to 1, indicating the port has    */
		/* been hit. offset is the port # we wish to update   */
		target->plist[(offset/8)] |= 1<<(offset%8); // voodoo
		/* increment the appropriate counters */
		target->prtcount++;
		target->parent->totalPrtcount++;

		if(target->parent->totalPrtcount > sdata.portThreshold)
		{
			if(target->parent->portsExceeded == FALSE) //new ps, alert!
			{
				SLog(p, 0, target->parent);
				SAlert(p, 0, target->parent);
				target->parent->portsExceeded = TRUE;
				dumpPacketStats(target->parent);
				FilterThis = TRUE;
			}
			else // old portscan, log the packet
			{
				SLog(p, 0, target->parent);
				FilterThis = TRUE;
			}
		}
}

/*********************************/
/* check to see if a port is set */
/*********************************/
inline int portIsSet(char *portlist, int offset)
{
	/* see comment in AddPort regarding this */
	return portlist[(offset/8)] & (1<<(offset%8));
}


/**************************************************/
/* Add a target to a portscanners target tree.    */
/* Called whenever a new target is touched by     */
/* a portscanner                                  */
/**************************************************/
void AddTarget(Portscanner *ps, Packet *p)
{
		struct in_addr tmp;
		Target *target;
		/* grab a node from the target pool */
		target = Target_borrow(); 
		/* fill in the target struct */
		target->target_ip = (u_int32_t)p->iph->ip_dst.s_addr;
		target->prtcount = 1;
		target->initial_time.tv_sec = p->pkth->ts.tv_sec;
		target->last_update_time.tv_sec = p->pkth->ts.tv_sec;
		target->parent = ps;
		
		Initportlist(target);  // zeros out the node's port list
		/* insert the new target node into the tree */
		(void)ubi_sptInsert(ps->targetRootPtr, (ubi_btNodePtr)target,
						(ubi_btNodePtr)target, NULL);

		/* update the target count */
		target->parent->totalTgtcount++;
		/* update the targets port list */
		AddPort(target, p->dp, p);
		/* check thresholds to see if this qualifies as a port scan */
		if(ps->totalTgtcount > sdata.tgtThreshold)
		{
			if(ps->targetsExceeded == FALSE) // if FALSE, then new portscan
			{
				tmp.s_addr = ps->scanner_ip;
				//printf("Portscanner %s # targets exceeded\n", inet_ntoa(tmp));
				//printf("Targets Exceeded\n");
				SLog(p, 0, ps); // log the packet
				SAlert(p, 0, ps); // generate an alert
				ps->targetsExceeded = TRUE; // we have now alerted
				dumpPacketStats(ps);
				FilterThis = TRUE;
			}
			else //alert has already been generated so log the packet
			{
				SLog(p, 0, ps);
				FilterThis = TRUE;
			}
		}
		return;
}

/************************************************************************/
/* Keeps some basic stats on each packet a portscanner sends before     */
/* he violates any thresholds.                                          */
/************************************************************************/
void addPacketStats(Portscanner *ps, Packet *p)
{
	scannerStat *pstat;

	pstat = (scannerStat *)malloc(sizeof(scannerStat)); //need to get rid of malloc
	pstat->timestamp.tv_sec = p->pkth->ts.tv_sec;
	pstat->timestamp.tv_usec = p->pkth->ts.tv_usec;
	pstat->src = p->iph->ip_src;
	pstat->dst = p->iph->ip_dst;
	pstat->proto = p->iph->ip_proto;
	if (p->sp)
	{
		pstat->sport = p->sp;
		pstat->dport = p->dp;
	}
	else if (p->icmph)
	{
		pstat->sport = p->icmph->type;
		pstat->dport = p->icmph->code;
	}
	pstat->len = ntohs(p->iph->ip_len);
	/* link to the rest of the stats for this Portscanner */
	pstat->next = ps->stat;
	ps->stat = pstat;
}

/******************************************************************/
/* logs the accumulated packet stats once a threshold is crossed  */
/******************************************************************/
void dumpPacketStats(Portscanner *ps)
{
	scannerStat *node;
	char msg[STD_BUF];
	char srcip[STD_BUF];
	char dstip[STD_BUF];
	char timestamp[TIMEBUF_SIZE];
	node = ps->stat;
	while (node->next != NULL)
	{
		// TODO: how should this look!!!
		ts_print((struct timeval *) &node->timestamp, timestamp);
		snprintf(srcip, STD_BUF, inet_ntoa(node->src));
		snprintf(dstip, STD_BUF, inet_ntoa(node->dst));
	    snprintf(msg, STD_BUF, "HISTORY time: %s src: %s dst: %s proto: %s sport: %u dport: %u dgmlen: %u\n", timestamp, srcip, dstip, protocol_names[node->proto], node->sport, node->dport, node->len);
		fwrite(msg, strlen(msg), 1, scanlog);
		fflush(scanlog);
		node = node->next;
	}
}

/*****************************************************************/
/* Adds a new portscanner to the portscan tree, builds a target  */
/* tree for this portscanner.                                    */
/*****************************************************************/
void AddPortScanner(Packet *p)
{
	Portscanner *ps;
	/* borrow a portscanner node from the portscanner node pool */
	ps = Portscanner_borrow();
	/* fill in the portscanner struct */
	ps->scanner_ip = (u_int32_t)p->iph->ip_src.s_addr;
    ps->last_update_time.tv_sec = p->pkth->ts.tv_sec;
    ps->initial_time.tv_sec = p->pkth->ts.tv_sec;
	ps->ssnid = ((StateRecord *)p->state)->session_id;
	ps->totalPrtcount = 1;
	ps->totalTgtcount = 1;
	ps->targetRootPtr = &ps->targetRoot;
	ps->portsExceeded = FALSE;
	ps->targetsExceeded = FALSE;
	/* create a new target tree for this portscanner */
	ubi_trInitTree(ps->targetRootPtr, 
			targetCompareFunc,
			0);
	
	/* Add the target to the target tree */
	AddTarget(ps, p);
    /* add this portscanner to the pruneList */
	UpdatePruneList(ps);
	/* get the stats for the initiating packet */
	addPacketStats(ps, p);
	/* add this scanner to the portscan tree */
	ubi_sptInsert(psRootPtr, (ubi_btNodePtr)ps,
					(ubi_btNodePtr)ps, NULL);
}

/***************************************************************/
/* Finds a session in the session tree and removes it          */
/***************************************************************/
void PruneSessionTree(u_int32_t ssnid)
{
	unsigned long tmp;
	unsigned long *returned;

	tmp = ssnid;
	returned = (unsigned long *)ubi_sptFind(sessionRootPtr, &tmp);
	if(returned==NULL) /* the session isn't in the tree */
	{
		return;
	}
	else /* nuke the session */
	{
		ubi_sptRemove(sessionRootPtr, (ubi_btNodePtr)returned);
	}
	free(returned);
	return;
}

/**************************************************************/
/* removes the oldest nodes from the pruneList, cleans up     */
/* their associated trees and returns nodes to their          */
/* respective pools. Used for a forced timeout.               */
/**************************************************************/
void TruncateScanList(plist_t *pruneList, int delnodes)
{
	
	int i;
	Portscanner *currentNode = NULL;

	/* empty list, get out */
	if(list_isempty(pruneList))
	{
#ifdef DEBUG
		printf("Empty list in PruneScanlist, returning from Truncate\n\n");
#endif
		return;  
	}

	//currentNode = list_last(pruneList);

	for(i=0; i<delnodes; ++i)
	{
		if(list_isempty(pruneList))
			return;

		currentNode = list_last(pruneList);
		/* Remove the last node from the pruneList */ 
		currentNode = list_delete(pruneList, currentNode);
		/* Remove the session from the session tree */
		PruneSessionTree(currentNode->ssnid);
		/* delete target tree for this portscanner (tmp) */
		DeleteTargetSpd(currentNode->targetRootPtr);
		/* remove currentNodeScanner from the port scanner tree */
		currentNode = (Portscanner *)ubi_sptRemove(psRootPtr, (ubi_btNodePtr)currentNode);
		/* return now unused currentNodeScanner to the node pool */
		Portscanner_return(currentNode);
	}
}
	
/****************************************************************/
/* Checks the pruneList for expired nodes, if it finds any it   */
/* will remove the associated trees and return nodes to their   */
/* node pools. This func runs everytime a new session is        */
/* established                                                  */
/****************************************************************/
void PruneScanlist(plist_t * pruneList, spp_timeval *watchPeriod,
				   spp_timeval *currentTime)
{
	Portscanner *currentNode;
	/* empty list, get out */
	if(list_isempty(pruneList))
	{
#ifdef DEBUG
		printf("Empty list in PruneScanlist, returning\n\n");
#endif
		return;  
	}
	currentNode = list_last(pruneList); // the last node is oldest
	while(currentNode)
	{
		if((currentTime->tv_sec - currentNode->last_update_time.tv_sec) > watchPeriod->tv_sec)
		{
			/* Remove the last node from the pruneList */ 
			currentNode = list_delete(pruneList, currentNode);
			/* Remove the session from the session tree */
			PruneSessionTree(currentNode->ssnid);
			/* delete target tree for this portscanner (tmp) */
			/* DeleteTargetSpd will return freed target nodes to pool */
			DeleteTargetSpd(currentNode->targetRootPtr);
			/* remove currentNodeScanner from the port scanner tree */
			currentNode = (Portscanner *)ubi_sptRemove(psRootPtr, (ubi_btNodePtr)currentNode);
			/* return now unused currentNodeScanner to the node pool */
			Portscanner_return(currentNode);
			/* setup for the next iteration */
			if (!list_isempty(pruneList))
				currentNode = list_last(pruneList);
			else
			{
				return;
			}
		}
		else
		{
			return;  /* linked list is sorted by last update, so */
			         /*	if any iteration fails the timeout       */
			         /*	then we are out of timed out sessions    */
		}
	}
}
			
/*********************************************************************/
/* Callback function used by splay trees to sort portscanner nodes   */
/*********************************************************************/
static int psCompareFunc(ubi_trItemPtr ItemPtr, ubi_trNodePtr NodePtr)
{
    Portscanner *nPs;
    Portscanner *iPs;

    nPs = (Portscanner *) NodePtr;
    iPs = (Portscanner *) ItemPtr;
    
    if(nPs->scanner_ip < iPs->scanner_ip) return 1;
    else if(nPs->scanner_ip > iPs->scanner_ip) return -1;

    return 0;
}


/*********************************************************************/
/* Callback function used by splay trees to sort target nodes        */
/*********************************************************************/
static int targetCompareFunc(ubi_trItemPtr ItemPtr, ubi_trNodePtr NodePtr)
{
    Target *nTarget;
    Target *iTarget;

    nTarget = (Target *) NodePtr;
    iTarget = (Target *) ItemPtr;

    if(nTarget->target_ip < iTarget->target_ip) return 1;
    else if(nTarget->target_ip > iTarget->target_ip) return -1;

    return 0;
}


/*********************************************************************/
/* Callback function used by splay trees to sort session nodes       */
/*********************************************************************/
static int ssnCompareFunc(ubi_trItemPtr ItemPtr, ubi_trNodePtr NodePtr)
{
    unsigned long *nSession;
    unsigned long *iSession;

    nSession = (unsigned long *) NodePtr;
    iSession = (unsigned long *) ItemPtr;

    if(nSession < iSession) return 1;
    else if(nSession > iSession) return -1;

    return 0;
}

/********************************************************************/
/* Generates a snort alert when a portscan is detected              */
/********************************************************************/
void SAlert(Packet *p, int scan_type, Portscanner *ps)
{
	Event event;
	char outstring[255];

	snprintf(outstring, 255, "spp_scanmunge: Portscan detected");
	SetEvent(&event, GENERATOR_SPP_SCAN2, SCAN_TYPE, 1, 0, 0, 0);
	CallAlertFuncs(p, outstring, NULL, &event);
	//ps->eid = event.event_id;
	ps->eid = 0;
}

/*******************************************************************/
/* Called for each packet of a portscan. Logs interesting packet   */
/* data to a text file                                             */
/*******************************************************************/
void SLog(Packet *p, int scan_type, Portscanner *ps)
{
	char msg[STD_BUF];
	char src[STD_BUF];
	char dst[STD_BUF];
	char timestamp[TIMEBUF_SIZE];
	char flagString[9];

	strncpy(src, inet_ntoa(p->iph->ip_src), STD_BUF);
	strncpy(dst, inet_ntoa(p->iph->ip_dst), STD_BUF);
	ts_print((struct timeval *) &p->pkth->ts, timestamp);
	//if (p->tcph->th_flags)
	if (p->tcph)
	{ 
		CreateTCPFlagString(p, flagString);
		snprintf(msg, STD_BUF, "%s TCP src: %s dst: %s sport: %u dport: %u tgts: %u ports: %u flags: %s eid: %u\n", timestamp, src, dst, p->sp, p->dp, ps->totalTgtcount, ps->totalPrtcount, flagString, ps->eid);
	}
	//else if(p->udph->uh_sport)
	else if(p->udph)
	{
		snprintf(msg, STD_BUF, "%s UDP src: %s dst: %s sport: %u dport: %u tgts: %u ports: %u eid: %u\n", timestamp, src, dst, p->sp, p->dp, ps->totalTgtcount, ps->totalPrtcount, ps->eid);
	}
	//else if(p->icmph->type)
	else if(p->icmph)
	{
		snprintf(msg, STD_BUF, "%s ICMP src: %s dst: %s type: %u code: %u tgts: %u eid: %u\n", timestamp, src, dst, p->icmph->type, p->icmph->code, ps->totalTgtcount, ps->eid);
	}

	fwrite(msg, strlen(msg), 1, scanlog);
	fflush(scanlog);
	
}

/*********************************************************************/
/* This is the main dude. Called by spp_conversation each time a new */
/* session is established.                                           */
/*********************************************************************/
void psWatch(Packet *p)
{
	Portscanner tmp;
	Portscanner *returned;
    Target tgt;
    Target	*rtgt;

#ifdef DEBUG
	printf("In PsWatch....\n");
#endif
	/* check to see if this guy is on the ignored list, if so bail */
	if(IsIgnored(p))
		return;

	/* search for this portscanner in the portscan tree */
	tmp.scanner_ip = (u_int32_t)p->iph->ip_src.s_addr;
	returned = (Portscanner *) ubi_sptFind(psRootPtr, (ubi_btItemPtr)&tmp);
#ifdef DEBUG
	printf("Found portscanner: %p\n", returned);
#endif

	if(returned == NULL)  /* we have a new potential scanner */
	{
		AddPortScanner(p);
	}
	else if(returned->ssnid == ((StateRecord *)p->state)->session_id)  /* session already logged, get out */
	{
        /* new activity so put his node on the front of the pruneList */
		UpdatePruneList(returned); 
		/* all done, bail */
		return;
	}
	else  /* Portscanner found, new session, update target/port */
	{
        returned->last_update_time.tv_sec = p->pkth->ts.tv_sec;

        tgt.target_ip = (u_int32_t)p->iph->ip_dst.s_addr;
		/* check to see if target has been hit before */
        rtgt = (Target *) ubi_sptFind((ubi_trRootPtr)returned->targetRootPtr, (ubi_btItemPtr)&tgt);

        if(rtgt == NULL) /* no such target in target tree, add him */
        {
            AddTarget(returned, p);  // AddTarget calls AddPort, so no need to call here
			if ((returned->targetsExceeded == FALSE) && (returned->portsExceeded == FALSE))
			{
        		addPacketStats(returned, p);
			}
		}
        else  /* target found in target tree */
        {
            if(!portIsSet(rtgt->plist, p->dp))  /* hasn't hit this port before */
            {
				/* update the port list for the target */
                AddPort(rtgt, p->dp, p);
				if ((returned->targetsExceeded == FALSE) && (returned->portsExceeded == FALSE))
				{
					addPacketStats(returned, p);
				}
            }
			/* move his node to the top of the pruneList */
			UpdatePruneList(returned);
        }
    }
	/* Cull any expired sessions out */
	PruneScanlist(pruneList, &sdata.timeout,  (spp_timeval *)&p->pkth->ts);
}

void SetupScan2(void)
{
        RegisterPreprocessor("scanmunge", Scan2Init);
}

/****************************************************/
/* Called at runtime to set everything up           */
/****************************************************/
void Scan2Init(u_char *args)
{
        struct timeval tv;
        struct timezone tz;
		extern conversationData cdata;

		if(cdata.isInitialized != 1)
		{
			FatalError("Please activate spp_conversation before"
					" trying to activate spp_scanmunge\n");
		}	

		ParseScanmungeArgs(args);
        gettimeofday(&tv, &tz);
		/* set up the portscanner tree */ 
	    (void)ubi_trInitTree(psRootPtr,
 							psCompareFunc,
        	                0);
		/* set up the session tree */
		(void)ubi_trInitTree(sessionRootPtr,
							 ssnCompareFunc,
							 0);
		/* set up the node pools */
		InitNodePools(sdata.psnodes, sdata.tgtnodes);
		/*set up the pruneList */
		pruneList = list_create(10000); //TODO should be an arg
		mdata.isInitialized = 1;
}
 
/* not using this right now, but it needs to used if using stream4 states */
void Scan2PreprocFunction(Packet *p)
{
		unsigned long tmp;
		unsigned long *returned;
		
		
        if (p->iph == NULL)
        {
                return;
        }
        switch(p->iph->ip_proto)
        {
                case IPPROTO_TCP:
                    if(p->tcph == NULL)
                        {
                    	    return;
                        }
 
					tmp = ((StateRecord *)p->state)->session_id;
					returned = (unsigned long *)ubi_trFind(sessionRootPtr, &tmp);
					
					if(returned == NULL) /* this is a new session */
					{
						psWatch(p);
					}
					else   /* session is already established */
					{
						return;
					}
                break;
 
                case IPPROTO_UDP: /* not doing udp yet */
                        /*if(p->udph == NULL)
                        {
                                return;
                        }
                        if (((unsigned long *)p->ssnptr)->ssnid)
                        {                                printf("p->ssnptr->ssnid: %i",  ((unsigned long *)p->ssnptr)->ssnid);
                        }*/
                        return;
 
                        break;

				case IPPROTO_ICMP: /*not doing icmp yet */
						return;
						break;
 
                default:
                        return;
                        break;
        }
}

/*******************************************/
/* checks to see if the pruneList is empty */
/*******************************************/
int list_isempty(plist_t *list)
{
	return list->nodecount == 0;
}

/**********************************************/
/* checks to see if a portscanner node is the */
/* first one on the pruneList                 */
/**********************************************/
int list_isfirst(plist_t *list, Portscanner *node)
{
	if (list->nilnode.pnext == node)
		return 1;
	return 0;
}

/*********************************************/
/* Grabs the last node off of the pruneList  */
/*********************************************/
Portscanner* list_last(plist_t *list)
{
	//printf("in list_last\n");
	if (list->nilnode.pprev == &list->nilnode)
		return NULL;
	return list->nilnode.pprev;
}

/*********************************************/
/* Removes a node from the pruneList         */
/*********************************************/
Portscanner* list_delete(plist_t *list, Portscanner *del)
{
	Portscanner *next = del->pnext;
	Portscanner *prev = del->pprev;

	prev->pnext = next;
	next->pprev = prev;

	list->nodecount--;
	del->pnext = del->pprev = NULL;

	return del;
}

/********************************************/
/* creates a pruneList                      */
/********************************************/
plist_t* list_create(int size)
{
	plist_t *new = malloc(sizeof *new);
	if (new) {
		new->nilnode.pnext = &new->nilnode;
		new->nilnode.pprev = &new->nilnode;
		new->nodecount = 0;
		new->maxcount = size;
	}
	return new;
}

/************************************************/
/* searches the pruneList for a particular node */
/************************************************/
int list_contains(plist_t *list, Portscanner *node)
{
	Portscanner *n = &list->nilnode;
	Portscanner *nil = &list->nilnode;

	if ((list->nodecount == 1) & (n==node))
		return 1;

	if ((list->nodecount == 1) & (n!=node))
		return 0;

	for (n = list->nilnode.pnext; n != nil; n = n->pnext) {
		if (node == n)
			return 1;
	}

	return 0;
}
	
/************************************************/
/* sticks the given node on the front of the    */
/* pruneList                                    */
/************************************************/
void list_prepend(plist_t *list, Portscanner *node)
{
	Portscanner *nil = &list->nilnode;
	Portscanner *nilnext = nil->pnext;

	/* the order looks like this:
	*	nil->nilnext  becomes
	*	nil->node->nilnext
	*/
	
	node->pprev = nil;     /* stick node right after first node */
	node->pnext = nilnext; 
	nilnext->pprev = node; 
	nil->pnext = node;
	list->nodecount++;
}
